/***************************************************************
 ************************* A TPP Project ***********************
 ******************** Visit http://tpp.dhs.org *****************
 ***************************************************************
dstring.cpp

Author(s): Paul Fleischer (polle@denmark.crosswinds.net)
Created: 18. July 1999

Copyright(c) 1999 The AKOS team.
All Rights Reserved.
****************************************************************/

#include "dstring.h"

//Constructor
dstring::dstring(char *stdvalue)
{
  isalloc = false;
  size = 0;

  if(stdvalue != NULL)
    {
      size = strlen(stdvalue)+1; //+1 because strlen doesn't count the \0 char
      data = (char*) malloc(size);
      strcpy(data, stdvalue);
      isalloc = true;
    }
}

//Destructor
dstring::~dstring()
{
  dealloc();
}

void
dstring::operator =(char *string)
{
  if(strlen(string) > 0)
  {
    dealloc();
    
    size = strlen(string)+1;
    
    alloc();
    
    strcpy(data, string);
  }
  else
  {
    size = 0;
    dealloc();
  }
}

void
dstring::operator =(int i)
{
  dealloc();

  itos(i);
}
 
void
dstring::operator =(double f)
{
  dealloc();

  ftos(f);
}

void
dstring::operator =(dstring ds)
{
  dealloc();

  size = strlen(ds.tochar());
  
  alloc();
  
  strcpy(data, ds.tochar());
}


void
dstring::operator +=(char *string)
{
  if(strlen(string) > 0)
  {
    char *tmpChar;
    int tmpSize;

    tmpSize = size + strlen(string);
    tmpChar = (char*) malloc(tmpSize);
    
    strcpy(tmpChar, data);
    
    strcat(tmpChar, string);
    
    size = tmpSize;
    
    dealloc();
    alloc();
    
    strcpy(data, tmpChar);
    free(tmpChar);
  }  
}

void
dstring::operator +=(int i)
{
  char *tmpChar;
  unsigned int digits=0;

  for(int l=i;l>1;digits++)
    l=div(l, 10).quot;

  if(digits == 0) digits = 1;
  if(i<1) digits++;

  tmpChar = (char*) malloc(size);
  strcpy(tmpChar, data);
  
  size += digits;

  dealloc();
  alloc();

  strcpy(data, tmpChar);
  free(tmpChar);

  tmpChar = (char*) malloc(digits);
  sprintf(tmpChar, "%i", i);
  strcat(data, tmpChar);

  free(tmpChar);
}

void
dstring::operator +=(double f)
{
  char *tmpChar;
  char tmp[200];

  sprintf(tmp, "%f", f);

  tmpChar = (char*) malloc(size);
  strcpy(tmpChar, data);
  
  size += strlen(tmp);

  dealloc();
  alloc();

  strcpy(data, tmpChar);
  free(tmpChar);

  strcat(data, tmp);

}

void
dstring::operator +=(char c)
{
  if(isalloc)
  {
    char *tmp;

    tmp = (char*)malloc(size);

    strcpy(tmp, data);

    size++;
    free(data);
    data = (char*)malloc(size);
    if(data == NULL)
    {
      printf("NOT ENOUGH MEMORY!!!!!!!!!!!!\n");
      _exit(1);
    }
    strcpy(data, tmp);

    data[size-2] = c;
    data[size-1] = 0;
   
    free(tmp);
  }
}

char *
dstring::tochar()
{
  if(isalloc)
    return data;
  else
    return 0;
}

int
dstring::toint()
{
  if(isalloc)
    return atoi(data);
  else
    return 0;
}

double
dstring::todouble()
{
  if(isalloc)
    return strtod(data, 0);
  else
    return 0;
}

//String manipulation functions

unsigned int
dstring::index(char c)
{
  if(isalloc)
    {
      unsigned int i=1;
      for(; (data[i-1] != 0) && (data[i-1] != c) ;i++);
      if(data[i-1] == 0)
	i = 0;
      return i;
    }
  else
    return 0;
}

void
dstring::cpy(dstring *dstString, unsigned int start, unsigned int end)
{
  if( (isalloc) && (end>start) )
    {
      unsigned int length = end-start;
      char *tmppoint;

      tmppoint = (char*)malloc(length+1);

      strncpy(tmppoint, data+(start-1), length);
      tmppoint[length] = 0;

      *dstString = tmppoint;

      free(tmppoint);
    }
  else
    *dstString = "";
}

unsigned int
dstring::getLength(void)
{
  if(isalloc)
    {
      return strlen(data)+1; //+1 because the dstring's first char. is 1, and not 0. 
    }
  else
    return 0;
}

//******************* Private functions *********************
void
dstring::itos(int i)
{
  if(i == 0)
    {
      char ret[] = "0";
      size = strlen(ret);
      
      alloc();

      strcpy(data, ret);
      return;
    }
  unsigned int digits=0;

  //Count the digits.. clumpsy way, but it works ;)
  for(int l=i;l>1;digits++)
    l=div(l, 10).quot;

  if(digits == 0) digits = 1;
  if(i<1) digits++;

  size = digits;

  alloc();

  sprintf(data, "%i", i);
}

void
dstring::ftos(double f)
{
  if(f == 0)
    {
      char ret[] = "0";
      size = strlen(ret);
      
      alloc();
      
      strcpy(data, ret);
      return;
    }

  //I know that this isn't the best way... but I see it as the only one ;)
  char ret[200];
  sprintf(ret, "%f", f);
  
  size = strlen(ret);

  alloc();

  sprintf(data, "%f", f);
}

void
dstring::dealloc()
{
  if(isalloc)
    {
      free(data);
    }
  isalloc = false;
}

void
dstring::alloc()
{
  dealloc();

  if(size > 0)
    {
      data = (char *)malloc(size);
  
      if(data != NULL)
	isalloc = true; 
    }
}

/*********************************************************************
 ******************************** darray *****************************
 *********************************************************************/

darray::darray(void)
{
  noOfItems = 0;
}

darray::~darray()
{
  if(noOfItems > 0)
    {
      for(unsigned int i=0;i<noOfItems;i++)
	delete items[i];
      free(items);
    }
}

//Resizes the array, right now it can only make the array larger
void
darray::resize(unsigned int newSize)
{
  dstring **tmpArray;

  if(noOfItems > 0)
    {
      tmpArray = (dstring**)malloc(noOfItems);
      
      for(unsigned int i=0;i<noOfItems;i++)
	{
	  tmpArray[i] = items[i];
	}
      
      free(items);

      items = (dstring**)malloc(newSize);

#ifdef VC
      for(i=0;i<noOfItems;i++)
#else
      for(unsigned int i=0;i<noOfItems;i++)
#endif
	{
	  items[i] = tmpArray[i];
	}
      free(tmpArray);
    }
  else
    items = (dstring**)malloc(newSize);

  items[noOfItems] = 0;
  noOfItems = newSize;
}

void
darray::setItem(unsigned int item, dstring *newItem)
{
  if(item+1 > noOfItems)
    {
      resize(noOfItems+1);
    }
  
  if(items[item] != NULL)
    delete items[item];

  items[item] = newItem;
}

dstring*
darray::getItem(unsigned int item)
{
  if(item+1 > noOfItems)
    return 0;
  else
    return items[item];
}

void
darray::delItem(unsigned int item)
{
  if(noOfItems > 0)
    {
      if(item == noOfItems-1)
	{
	  items = (dstring**)realloc(items, noOfItems-1);
	  
	}
      else
	{
	  for(unsigned int i=item+1;i<noOfItems;i++)
	    items[i-1] = items[i];
	  
	  items = (dstring**)realloc(items, noOfItems-1);
	}
      noOfItems--;
    }
}

unsigned int
darray::getSize(void)
{
  return noOfItems;
}
