/***************************************************************
 ************************* A TPP Project ***********************
 ******************** Visit http://tpp.dhs.org *****************
 ***************************************************************
main.cpp

Author(s): Paul Fleischer (polle@denmark.crosswinds.net)
Created: 20. June 1999

Copyright(c) 1999 The AKOS team.
All Rights Reserved.
****************************************************************/

#include "main.h"

aobject globalobjects("global");

unsigned int typeAssignment(CODE_BLOCK *code, unsigned int chpos,
			    unsigned int *lineNumber, aobject *object);
unsigned int firstSecondArg(CODE_BLOCK *code, unsigned int chpos,
			    unsigned int *lineNumber, dstring *firstArg, 
			    dstring *secondArg);
unsigned int defineObject(CODE_BLOCK *code, aobject *object, dstring *firstArg,
			  dstring *secondArg, unsigned int chpos, 
			  unsigned int *lineNumber);
unsigned int processLine(CODE_BLOCK *code, unsigned int i, 
			 unsigned int *lineNumber, aobject *object, 
			 unsigned char *lineType=0);
unsigned int assignObject(CODE_BLOCK *code, aobject *object, dstring *firstArg,
			  unsigned int chpos, unsigned int *lineNumber);
void doAssign(CODE_BLOCK *code, aobject *object, dstring *firstArg, 
	      dstring *value, unsigned int *lineNumber);
unsigned int specialMethod(CODE_BLOCK *code, unsigned int chpos,
			   unsigned int *lineNumber, aobject *object);
void specialRun(CODE_BLOCK *code, dstring *firstArg, avalue *output, 
		aobject *object, unsigned int lineNumber);


//Test main function.
int
main(int args, char **argv)
{
  CODE_BLOCK testCode;
  char *codeString = "type point = (\n\
  int(2) x\n\
  int y\n\
  method reset = (int(2) newX, int(2) newY, x = newX, y = newY)\n\
  )\n\
  point = 23 \"hejsa\"\n\
  point.reset()\n\
\n";

  printf("AHLL version %s\nCopyright (c) 1999 The Programming Place\n\n\n", VERSION);

  //Setup global objects
  globalobjects.spawnChild("type"); //type is a template with no children
  globalobjects.spawnChild("int");  //Actually the same, but it's value is used
  globalobjects.getChild("int")->setData("0");
  globalobjects.spawnChild("debug");   //The debug object.
  globalobjects.spawnChild("method"); //The method object.

  //Setup a test CODE_BLOCK

  //Setup the local top-object
  testCode.topobject = new aobject("Test_Code", &globalobjects);

  testCode.name = new dstring("Test_Code");
  testCode.data = (char*)malloc(strlen(codeString));
  strcpy(testCode.data, codeString);
  testCode.dataSize = strlen(testCode.data);

  //Now interpret the test CODE_BLOCK
  analyseString(&testCode);

  #ifdef DEBUG
  fprintf(stderr, "end reached\n");
  #endif

  free(testCode.data);
  return 0;
}

/**********************************************************************
-----------------------------------------------------------------------
void
emitError(char *module, int line, char *messgae)
-----------------------------------------------------------------------
Info: Sends error messages to the debugging system.
      Currently this function does only print a error message to stdout

In: 'module'; The name of module, were the error has occured.
    'line'; The line were the error has occured.
    'message'; The error message to send.
    
Out: <NONE>

Return: void
***********************************************************************/
void
emitError(char *module, int line, char *message)
{
  printf("\n*********************************\n");
  printf("PRE Monitoring System: ERROR:\n");
  printf("%s: line %i: %s\n", module, line, message);
  printf("*********************************\n");
}

/**********************************************************************
-----------------------------------------------------------------------
void
emitWarning(char *module, int line, char *messgae)
-----------------------------------------------------------------------
Info: Sends warning messages to the debugging system.
      Currently this function does only print a warning message to stderr

In: 'module'; The name of module, were the error has occured.
    'line'; The line were the error has occured.
    'message'; The error message to send.
    
Out: <NONE>

Return: void
***********************************************************************/
void
emitWarning(char *module, int line, char *message)
{
  fprintf(stderr, "\n****************************\n");
  fprintf(stderr, "PRE Monitoring System: WARNING:\n");
  fprintf(stderr, "%s: line %i: %s\n", module, line, message);
  fprintf(stderr, "****************************\n");
}

/**********************************************************************
-----------------------------------------------------------------------
bool
analyseString(CODE_BLOCK *code)
-----------------------------------------------------------------------
Info: This function does the actual interpreting

In: 'code' the CODE_BLOCK to be interpreted
    
Out: <NONE>

Return: void
***********************************************************************/
void
analyseString(CODE_BLOCK *code)
{
  unsigned int lineNumber=1;
  unsigned int i=0;

  printf("codeblock:\n%s\n", code->data);

  //Main loop
  for(;code->data[i] != 0;i++)
    {
      i = processLine(code, i, &lineNumber, code->topobject);
      if(i == 0) break;
      if(code->data[i] == '\n')
	lineNumber++;
    }

#ifdef DEBUG
  printf("Dumbing top-level objects:\n");
#ifdef VC
  for(i=0;i<code->topobject->getNoOfChilds();i++)
#else
  for(unsigned int i=0;i<code->topobject->getNoOfChilds();i++)
#endif
    {
      printf("%s\n", code->topobject->getChild(i+1)->getName()->tochar());
    }
  printf("Dumbing children of 'point':\n");

#ifdef VC
 for(i=0;i<code->topobject->getChild("point")->getNoOfChilds();i++)
#else
  for(unsigned int i=0;i<code->topobject->getChild("point")->getNoOfChilds();i++)
#endif
    if(code->topobject->getChild("point")->getChild(i+1)->getData() != NULL)
      printf("%s %s = %s\n", 
	     code->topobject->getChild("point")->getChild(i+1)->getType()->getName()->tochar(), 
	     code->topobject->getChild("point")->getChild(i+1)->getName()->tochar(), 
	     code->topobject->getChild("point")->getChild(i+1)->getData()->tochar());
#endif

}

/***********************************************************************
----------------------------------------------------------------------
unsigned int 
typeAssignment(CODE_BLOCK *code, unsigned int chpos,
               unsigned int &lineNumber, aobject *object)
----------------------------------------------------------------------
Info:  This function handles the optional third argument('=')
       after a object declaration.

In:    'code' is the CODE_BLOCK that is being interpreted
       'chpos' is the character counter.
       'lineNumber' is a pointer to the lineNumber counter, it is changed
       only if a third argument is present.
       'object' the object which is just declared, this is the one we
       are assigning sub-objects too.

Out:   'lineNumber' contains the new lineNumber if a third argument did
       exist, else it is unchanged.

Return: unsigned int, the new new chpos in case there exists a third argument,
        else it returns the same chpos as it was called with.
**************************************************************************/
unsigned int 
typeAssignment(CODE_BLOCK *code, unsigned int chpos, unsigned int *lineNumber, 
	       aobject *object)
{
  unsigned int line = *lineNumber;
  unsigned int newPos = chpos;
  bool thirdArg = false;

  //Start by skipping spaces, but not newlines
  for(; (code->data[newPos] != 0) &&
	(code->data[newPos] == ' ')
	;newPos++);

  if(code->data[newPos] == '=')
    {
      thirdArg = true;
      newPos++;
    }

  if(!thirdArg)
    return chpos;

  //Skip spaces AND newlines again
  for(; (code->data[newPos] != 0) &&
	((code->data[newPos] == ' ') ||
	 (code->data[newPos] == '\n'))
	;newPos++)
    if(code->data[newPos] == '\n') line++;

  //  newPos++;

  //Now check if the next character is a '('...
  if(code->data[newPos] == '(')
    {
      //...if yes, then it starts a definition of the obeject.
      //Run a loop which defines all the found objects as sub-objects
      //until it encounters a ')', which means the-end-of-definition
      newPos++;
      if(code->data[newPos] == '\n')
	line++;

      for(; (code->data[newPos] != 0)
	    ;newPos++)
	{
	  newPos = processLine(code, newPos, &line, object);
	  if(newPos == 0)
	    return 0;
	  if(code->data[newPos] == '\n')
	    line++;
	  if(code->data[newPos] == ')')
	    break;
	}
    }
  newPos++;
  if(code->data[newPos] == '\n')
    {
      line++;
    }
  *lineNumber = line;
  return newPos;
}


/***********************************************************************
----------------------------------------------------------------------
unsigned int 
firstSecondArg(CODE_BLOCK *code, unsigned int chpos,
               unsigned int &lineNumber,
	       dstring *firstArg, dstring *secondArg)
----------------------------------------------------------------------
Info:  This function finds the first and second argument, and returns them
       as dstrings.

In:    'code' is the CODE_BLOCK that is being interpreted
       'chpos' is the character counter.
       'lineNumber' is a pointer to the lineNumber counter, it is changed
       only if a third argument is present.

Out:   'lineNumber' will contain the new lineNumber
       'firstArg' will be set to the first argument
       'secondArg' will be set to the second argument
Return: unsigned int, the new new chpos in case there exists a third argument,
        else it returns the same chpos as it was called with.
**************************************************************************/
unsigned int 
firstSecondArg(CODE_BLOCK *code, unsigned int chpos,
	       unsigned int *lineNumber,
	       dstring *firstArg, dstring *secondArg)
{
  //Skip all spaces, and newlines
  for(;(code->data[chpos] != 0) && 
	((code->data[chpos] == ' ') ||
	(code->data[chpos] == '\n'))
	;chpos++)
    if(code->data[chpos] == '\n') 
      (void)*lineNumber++;
  
  //reached end? If yes, return 0
  if(code->data[chpos] == 0)
    return 0;
  
  //Now read the first command, spaces, line seperators(',' and newlines)
  //'(' and ')' are also counted, and as soon as there is are an equal amount
  // of '(' and ')' the loop stops.
  /*The following are special characters:
    '=' are not allowed in object names. It breaks the reading loop, ie. a '='
    seperates first and second argument.
  */
  int pcount=1;
  for(;(code->data[chpos] != 0) &&
	!(code->data[chpos] == ' ' && pcount < 2) &&
	(code->data[chpos] != '\n') &&
	(code->data[chpos] != ',') &&
	(code->data[chpos] != '=')
	;chpos++)
    {
      *firstArg += code->data[chpos];
      
      if(code->data[chpos] == '(')
	pcount++;
      if(code->data[chpos] == ')')
	pcount--;
    }

  //A ')' ends an expression, therefore return.
  if(code->data[chpos-1] == ')' && pcount==0)
    {
      chpos--;
      return chpos;
    }

  if(code->data[chpos] == '\n')
    {
      (void)*lineNumber++;
    }

  /*If the ending character is a line seperator(',' and newline) stop searching
   for a second argument.*/
  if(code->data[chpos] == '\n' || code->data[chpos] == ',')
    return chpos;

  //Now comes an important part, what comes after the first argument?
  
  //But first we need to skip spaces
  for(;(code->data[chpos] != 0) && 
	((code->data[chpos] == ' '))
	;chpos++)
    ;
      
  
  if(code->data[chpos] == 0)
    return 0;


  if(code->data[chpos] == '\n')
    {
      (void)*lineNumber++;
    }

  if(code->data[chpos] == '\n' || code->data[chpos] == ',')
    return chpos;


  //Read the "second argument" (stop when encountering a line seperator)
  //(Description of line seperator is above)
  /*There are a few special cases of the second-argument loop:
    If a '=' is encountered the loop is breaked right away. This is done
    because the '=' character has to be the only character in the second
    argument for a assignment to work.
   */
  for(;(code->data[chpos] != 0) &&
	(code->data[chpos] != ' ') &&
	(code->data[chpos] != '\n') &&
	(code->data[chpos] != ',') &&
	(code->data[chpos] != ')')
	;chpos++)
    {
      *secondArg += code->data[chpos];
      if(code->data[chpos] == '=')
	break;
    }

  if(code->data[chpos] == '=')
    chpos++;

  if(code->data[chpos] == '\n')
    (void)*lineNumber++;

  return chpos;
}


/***********************************************************************
----------------------------------------------------------------------
unsigned int
defineObject(CODE_BLOCK *code, aobject *object, dstring *firstArg,
	     dstring *secondArg, unsigned int chpos, unsigned int *lineNumber)
----------------------------------------------------------------------
Info:  Handles object definitions/declaration. A object definition/declaration
       looks like this:
       <type> <object-name> [= (<children>)]
       
       This clones the <type> object to <object-name>. 
       The last part is optional

In:    'code' is the CODE_BLOCK that is being interpreted
       'object' is the parent object, ie. the object(s) parent.
       'firstArg' is the first Argument (the <type> above)
       'secondArg' is the second Argumyn (<object-name> above)
       'chpos' is the character position counter.
       'lineNumber' is a pointer to the line number counter.

Out:   'lineNumber' will contain the new line number.

Return: unsigned int, the new character position.
**************************************************************************/
unsigned int
defineObject(CODE_BLOCK *code, aobject *object, dstring *firstArg,
	     dstring *secondArg, unsigned int chpos, unsigned int *lineNumber)
{
  aobject *tmpobject;
  aobject *curobject;
  bool invalidName = false;
  unsigned int i = chpos;
  bool assignValue = false;
  dstring value("");
  dstring tmpString("");

  /*Check the firstArg to see if there is a '('.
    If yes, then the object has to asigned the value between the '(' and ')'.
    This loop also removes the () part from the firstArg string.
   */
  for(unsigned int l=0; (l<strlen(firstArg->tochar())) && 
	firstArg->tochar()[l] != ')'
	;l++)
    {
      if(firstArg->tochar()[l] == '(')
	{
	  l++;
	  assignValue = true;
	}

      if(!assignValue)
	tmpString += firstArg->tochar()[l];
      else
	value += firstArg->tochar()[l];
    }

  *firstArg = tmpString.tochar();

  tmpobject = object->lookupObject(firstArg->tochar());
  if(object->getChild(secondArg->tochar()) != NULL)
    invalidName = true;
  
  //If the returned object is NULL, then the template doesn't exist
  if( (tmpobject == NULL) || (invalidName))
    {
      dstring message("");
      if(tmpobject == NULL)
	{
	  message = "Object '";
	  message += firstArg->tochar();
	  message += "' does not exist";
#ifdef DEBUG
	  message += "(defineObject)";
#endif
	  emitError(code->name->tochar(), *lineNumber, message.tochar());
	}
      
      if(invalidName)
	{
	  message = "An object with the name '";
	  message += secondArg->tochar();
	  message += "' does allready exist";
	  emitError(code->name->tochar(), *lineNumber, message.tochar());
	}
    }
  else
    {
      curobject = object->spawnChild(tmpobject, secondArg->tochar());
#ifdef DEBUG
      printf("spawning new child!\n");
      dstring tmpString("");
      curobject->getFullName(&tmpString);
      printf("(1): \"%s\"\n", tmpString.tochar());
      tmpString = " ";
      object->getFullName(&tmpString);
      printf("(2): \"%s\"\n", tmpString.tochar());
#endif
      if(strcmp(firstArg->tochar(), "method") == 0)
	i = specialMethod(code, i, lineNumber, curobject);
      else
	i = typeAssignment(code, i, lineNumber, curobject);
    }

  if(assignValue)
    {
      doAssign(code, object, secondArg, &value, lineNumber);
      dstring tmpString("");
      curobject->getFullName(&tmpString);
      printf("Assign value: %s = %s\n", tmpString.tochar(), value.tochar());
      printf("The new object has value: \"%s\"\n", curobject->getData()->tochar());
    }

  return i;
}


/***********************************************************************
----------------------------------------------------------------------
unsigned int
processLine(CODE_BLOCK *code, unsigned int i, unsigned int *lineNumber,
	    aobject *object, unsigned char *lineType)
----------------------------------------------------------------------
Info:  proccesses(interprets) a "line" of a code block.
       

In:    'code' is the CODE_BLOCK that is being interpreted
       'i' is the character counter.
       'lineNumber' is the line number counter, it will be changed, as
       newline chars are encountered.
       'object' is the parent object of the current proccess, or in other words
       the object that gets new children when a objects is defined in the code.

Out:   'lineNumber' will contain the new lineNumber
       'lineType' will contain a number representing the type of the line.
       if lineType is NULL when calling, nothing will happen.

Return: unsigned int, returns the new character counter.
**************************************************************************/
unsigned int
processLine(CODE_BLOCK *code, unsigned int i, unsigned int *lineNumber,
	    aobject *object, unsigned char *lineType)
{
  dstring firstArg("");
  dstring secondArg("");
  bool commandCall = false;
  bool assignment = false;
  bool unknown = false;
  bool retValue = false;

  i = firstSecondArg(code, i, lineNumber, &firstArg, &secondArg);

#ifdef DEBUG
  printf("Line %i (%s)\n1. arg: '%s'\n", *lineNumber, code->name->tochar(), firstArg.tochar());
  printf("2. arg: '%s'\n", secondArg.tochar());
#endif  

  //If firstSecondArg returns 0, then it has encountered a '\0' char
  if(i == 0)
    return 0;
  
  //If there is no second argument then we guess that it's a unnamed cloneing.
  //This is the same as running the method (formerly done by use of the run
  //keyword).
  //Notice, that if the first argument starts with a ')' the line is NOT a
  //command, this is because ')' characters normally are alone on lines.
  if( (secondArg.tochar()[0] == 0) && 
      (firstArg.tochar()[0] != 0) &&
      (firstArg.tochar()[0] != ')'))
    {
      commandCall = true;
      if(lineType != NULL)
	*lineType = 1;
    }


  //Handle the 'return' keyword
  //The return value can either be a aobject, or a value.
  //If it's an aobject, a pointer to it is returned.
  //If it's a value a new object with that value is created.
  if(strcmp(firstArg.tochar(), "return") == 0)
    {
      retValue = true;
  
      if(object->lookupObject(secondArg.tochar()) != NULL)
	code->returnValue = object->lookupObject(secondArg.tochar());
      else
	{
	  code->returnValue = new aobject("VALUE");
	  code->returnValue->setData(secondArg.tochar());
	}	
    }

  //If the first character of the second argument is a '=', then it's
  //probablt an assignment
  if(secondArg.tochar()[0] == '=')
    {
      assignment = true;
      if(lineType != NULL)
	*lineType = 2;
    }

  //If there is no second argument, the type of the command is currently
  //unknown, therefor ignore it
  if(secondArg.tochar()[0] == 0)
    {
      unknown = true;
      if(lineType != NULL)
	*lineType = 10;
    }

  //If it's neither an assignment, a return value nor a command, it's probably
  //a definition
  if(!commandCall && !assignment && !unknown && !retValue)
    {
      if(lineType != NULL)
	*lineType = 3;
      i = defineObject(code, object, &firstArg, &secondArg, i, lineNumber);
      if(i == 0)
	return 0;
    }

  //Object assignment
  if(assignment)
    {
      i = assignObject(code, object, &firstArg, i, lineNumber);
      if(i == 0)
	return 0;
    }

  if(commandCall)
    {
      specialRun(code, &firstArg, 0, object, *lineNumber);
    }

  if(code->data[i] == '\n')
    (void)*lineNumber++;

  return i;
}

//assignObject is the interface to doAssign
//assignObject gets the value list from the codedata.
unsigned int
assignObject(CODE_BLOCK *code, aobject *object, dstring *firstArg,
	     unsigned int chpos, unsigned int *lineNumber)
{
  dstring value("");

  //Skip spaces
  for(; (code->data[chpos] != 0) &&
	(code->data[chpos] == ' ')
	;chpos++);

  if(code->data[chpos] == 0)
    return 0;

  for(; (code->data[chpos] != 0) &&
	//(code->data[chpos] != ' ') &&
	(code->data[chpos] != ')') &&
	(code->data[chpos] != '\n') &&
	(code->data[chpos] != ',')
	;chpos++)
    value += code->data[chpos];

#ifdef DEBUG
  printf("firstArg: %s\n", firstArg->tochar());
  printf("value: %s\n", value.tochar());
#endif

  doAssign(code, object, firstArg, &value, lineNumber);

  if(code->data[chpos] == '\n')
    (void)*lineNumber++;

  return chpos;
}

/*
  doAssign does the actual assignment of objects
*/
void
doAssign(CODE_BLOCK *code, aobject *object, dstring *firstArg, dstring *value,
	 unsigned int *lineNumber)
{
  //This only works as long as firstArg in unchanged
  char *first = firstArg->tochar();

  bool stopLoop = false;
  aobject *firstObj = object;
  unsigned int noOfArgs = 0;

  /*
    Now go through any subobjects (like. parent.child.child), and get the
    correct object (= the last one).
    I know that the aobject class has a function for it (getSubObject), but 
    this is made before the getSubObject function. I will remove it when
    I feel like it ;-)
   */
  for(int i=0; !stopLoop; i++)
    {
      dstring obj("");
      for(; (first[i] != 0) && (first[i] != '.');i++)
	obj += first[i];
      
      firstObj = firstObj->lookupObject(obj.tochar());

      if(firstObj == NULL)
	{
	  dstring message("Object ");
	  message += obj.tochar();
	  message += " does not exist";
	  emitError(code->name->tochar(), *lineNumber, message.tochar());
	}

      if(first[i] == 0)
	stopLoop = true;
    }

  char *second = value->tochar();
  stopLoop = false;
  darray arguments;

  /*
    Read the assign argument list (seperated by spaces) and put the arguments
    into a darray variable (arguments)
   */
#ifdef VC
  for(i=0; !stopLoop; i++)
#else
  for(int i=0; !stopLoop; i++)
#endif
    {
      dstring argument("");
      bool inString = false;

      for(; !(second[i] == ' ' && !inString) && (second[i] != 0); i++)
	{
	  argument += second[i];
	  if(second[i] == '"')
	    {
	      if(inString)
		inString = false;
	      else
		inString = true;
	    }
	}

      if(second[i] == 0)
	stopLoop = true;

      arguments.setItem(noOfArgs, new dstring(argument.tochar()));
      
      noOfArgs++;
    }

  if(firstObj != NULL)
    {  
      /*
	If firstObj has children then...
       */
      if(firstObj->getNoOfChilds() > 0)
	{
	  /*
	    ...assign the children the values in the argument list.
	    ie: the first child will get the value of the first argument,
	    the second child will get the value of the second argument, etc.
	   */
	  bool runLoop = true;
	  for(int i=0;runLoop;i++)
	    {
	      aobject *tempobj;
	      
	      tempobj = firstObj->getChild(i+1);
	      
	      if((tempobj != NULL) && (arguments.getItem(i) != NULL))
		{
		  aobject *assignObj = object->getSubObject(arguments.getItem(i)->tochar());
		  /*
		    If assignObj exists(if the value of the argument is an
		    object) and if assignObj has a value...
		   */
		  if( (assignObj != NULL) && 
		      (assignObj->getDataPointer() != NULL))
		    {
		      /*
			...then assign it's value to the child pointed to by
			tempobj...
		       */
		      tempobj->setData( assignObj->getDataPointer());
#ifdef DEBUG	
		      printf("Setting %s to '%s'\n", tempobj->getName()->tochar()
			     , assignObj->getData()->tochar());
#endif
		    }	  
		  else
		    {		
		      /*
			...If assignObj does not exist(the value is of the
			argument is NOT an object) or if assignObj does not
			have any value, then assign the value of the argument
			(not assignObj) to the object pointed to by tempobj.
		       */
		      tempobj->setData(arguments.getItem(i)->tochar());
#ifdef DEBUG
		      printf("Setting %s to '%s'\n", tempobj->getName()->tochar(), arguments.getItem(i)->tochar());
#endif
		    }
		}
	      else
		runLoop = false;
	      
	    }
	}
      /*
	If firstObj does NOT have any children:
       */
      else
	{
	  if(firstObj != NULL)
	    {
	      aobject *tempObj = object->getSubObject(arguments.getItem(0)->tochar());
	      /*
		If tempObj exists(ie. the value of the first argument is an
		object)...
	       */
	      if( tempObj != NULL)
		{
		  /*
		    ...then assign the value of tempObj to firstObj
		   */
		  dstring fullName("");
		  tempObj->getFullName(&fullName);
		  firstObj->setData( tempObj->getDataPointer());
#ifdef DEBUG
		  printf("%s = %s(%s)\n", firstObj->getName()->tochar(), fullName.tochar(), tempObj->getData()->tochar());
#endif DEBUG
		}
	      /*
		If tempObj does NOT exist(ie. the value of the first argument
		is NOT an object)...
	       */
	      else	
		/*
		  ...then assign the value of the first arguement to firstObj.
		 */
		firstObj->setData(arguments.getItem(0)->tochar());
	    }
	}
    }
}

/*
  specialMethod is called by defineObject when the method object is cloned.
 */

unsigned int 
specialMethod(CODE_BLOCK *code, unsigned int chpos,
	      unsigned int *lineNumber, aobject *object)
{
  unsigned int newPos = chpos;
  unsigned int line = *lineNumber;


  //Start by skipping spaces, but not newlines
  for(; (code->data[newPos] != 0) &&
	(code->data[newPos] == ' ')
	;newPos++);

  if(code->data[newPos] != '=')
    {
      dstring message("'=' expected, but '");
      message += code->data[newPos];
      message += "' found!";
      emitError(code->name->tochar(), *lineNumber, message.tochar());
      return chpos;
    }

  if(code->data[newPos] == '\n' )
    line++;

  newPos++;

  //Skip spaces AND newlines again
  for(; (code->data[newPos] != 0) &&
	((code->data[newPos] == ' ') ||
	 (code->data[newPos] == '\n'))
	;newPos++)
    if(code->data[newPos] == '\n') 
      line++;

  //In case that the line is invalid, ie. no '(', return.
  if(code->data[newPos] != '(')
    {
      dstring message("'(' expected, but '");
      message += code->data[newPos];
      message += "' found!";
      emitError(code->name->tochar(), *lineNumber, message.tochar());
      return newPos;
    }

  //Now copy everything between the starting '(' and the ending ')'
  unsigned int pcount=1;
  dstring value("");

  newPos++;

  for(; (code->data[newPos] != 0) &&
	(pcount > 0)
	;newPos++)
    {
      if(code->data[newPos] == '(')
	pcount++;
      if(code->data[newPos] == ')')
	pcount--;

      if(code->data[newPos] == '\n')
	line++;

      if(pcount > 0)
	value += code->data[newPos];
    }

  value += '\n';

  *lineNumber = line;

  object->setData(&value);

#ifdef DEBUG
  printf("(0): \"%s\"\n", object->getData()->tochar());
#endif DEBUG

  return newPos;
}

void
specialRun(CODE_BLOCK *code, dstring *firstArg, aobject *output, 
	   aobject *object, unsigned int lineNumber)
{
  dstring secondArg("");
  unsigned int newPos=0;
  darray arguments;
  unsigned int noOfArgs=0;

#ifdef DEBUG
  printf("Argument to run: %s\n", firstArg->tochar());
#endif

  //The second argument is the name of the method, it's the name until a '('
  //or a space is encountered, the rest of the string(firstArg) are the 
  //arguments send to the method.
  for(; (firstArg->tochar()[newPos] != 0) &&
	(firstArg->tochar()[newPos] != '(')
	; newPos++)
    {
      secondArg += firstArg->tochar()[newPos];
    }

  if(firstArg->tochar()[newPos] == '(')
    newPos++;

  aobject *funcObject = code->topobject->getSubObject(secondArg.tochar());

  if(funcObject == NULL)
    {
      dstring message("Object '");
      message += secondArg.tochar();
      message += "' does not exist!";
      emitError(code->name->tochar(), lineNumber, message.tochar());
      return;
    }

  if(strcmp(funcObject->getType()->getName()->tochar(), "method") != 0)
    {
      dstring message("Object '");
      message += funcObject->getName()->tochar();
      message += "' is not a method";
      emitWarning(code->name->tochar(), lineNumber, message.tochar());
      return;
    }
  
  //Now get the arguments
  bool stopLoop = false;
  for(; !stopLoop; newPos++)
    {
      dstring argument("");
      bool inString = false;

      //Skip spaces
      for(; firstArg->tochar()[newPos] == ' ';newPos++);

      //Read argument
      for(; !(firstArg->tochar()[newPos] == ' ' && !inString) &&
	    (firstArg->tochar()[newPos] != 0)
	    ;newPos++)
	{
	  if( !((firstArg->tochar()[newPos] == ')') && 
		(firstArg->tochar()[newPos+1] == 0)) )
	    argument += firstArg->tochar()[newPos];
	  if(firstArg->tochar()[newPos] == '"')
	    {
	      if(inString)
		inString = false;
	      else
		inString = true;
	    }
	}
      
      if(firstArg->tochar()[newPos] == 0)
	stopLoop = true;

#ifdef DEBUG
      printf("%i. '%s'\n", noOfArgs, argument.tochar());
#endif

      if(argument.tochar()[0] != 0)
	arguments.setItem(noOfArgs, new dstring(argument.tochar()));

      noOfArgs++;
    }

  //Setup a new code object for the method
  CODE_BLOCK subCode;
      
  subCode.name = new dstring(secondArg.tochar());

  //The data is the value(data) of the function object 
  subCode.data = (char*)malloc(strlen(funcObject->getData()->tochar()));
  strcpy(subCode.data, funcObject->getData()->tochar());
  subCode.dataSize = strlen(funcObject->getData()->tochar());
  subCode.returnValue = NULL;

  subCode.topobject = funcObject;

  //Prepare to run the main loop
  unsigned int line = 1;
  unsigned int i=0;
  unsigned int count=0;

  //Start processing the subCode, until the lineType is not 3 (cloning).
  //Assign the arguments to the firsts sub-objects declared
   for(;subCode.data[i] != 0;i++)
    {
      unsigned char lineType=0;
      i = processLine(&subCode, i, &line, subCode.topobject, &lineType);
      if(lineType == 3)
	{
	  aobject *tmpobject = subCode.topobject->getChild(count+1);
	  if( (tmpobject == NULL) || (arguments.getItem(count) == NULL))
	    break;

	  dstring name("");
	  tmpobject->getFullName(&name);

#ifdef DEBUG
	  printf("%s = %s\n", name.tochar(), arguments.getItem(count)->tochar());
#endif

	  tmpobject->setData(arguments.getItem(count)->tochar());
	  count++;
	}
      else
	break;
      
      if(i == 0) break;
      if(subCode.data[i] == '\n')
	line++;
    } 
  
  if(i > 0)
    {
      for(;subCode.data[i] != 0;i++)
	{
	  i = processLine(&subCode, i, &line, subCode.topobject);
	  if(i == 0) break;
	  if(subCode.data[i] == '\n')
	    line++;
	}  
    }

#ifdef DEBUG
  if(subCode.returnValue != NULL)
    {
      if(subCode.returnValue->getParent() == NULL)
	{
	  printf("Value: %s\n", subCode.returnValue->getData()->tochar());
	}
      else
	printf("Object: %s\n", subCode.returnValue->getName()->tochar());
    }
#endif

  if(output != NULL)
    {
      if(subCode.returnValue != NULL)
	{
	  output = new aobject(returnValue, subCode.name->tochar());
	}
      else output = 0;
    }
  
  free(subCode.data);
}
